/***************************************************************************************
*
* Block_FIR.ASM
*
* Accompaniment for:
* Guide for Porting Code from SHARC DSPs to TigerSHARC Processors (EE-241)
*
* Written by Andrew Caldwell and Maikel Kokaly-Bannourah 
*	
* June 2004 rev 1.0 
*
* This modified version of the Block FIR code shows a one-to-one conversion (optimization 1)
* of the Block FIR SHARC code to TigerSHARC. 
*
***************************************************************************************/

/***********************************************************************

This is an example of a C callable floating point block FIR.
The example is for 64 taps with 128 samples and is provided to show
how the techniques detailed in the accompanying application note can be
applied to this example to convert it for operation on the TigerSHARC
family of processors.

Some of the macros for saving and restoring to and from the stack
have been added simply to show how the TigerSHARC macros provided with
the accompanying application note can be implemented without much
alteration of the code.

Optimizations performed in this example are:

1) Remove some redundant instructions

2) Create a new pointer that points to the place in the dline to store the new sample

3) Convert the two single loads required to load samples from the delay line
   so that a quad load is performed. This way the DAB can be used to allow for
   non-aligned acceses. The dline loads will require registers yxr3:0.
   For this reason I have modified the registers used for calculating and
   loading the number of loop iterations required. You also need to
   recalculate the number of times the inner loop is executed as we now require
   4 SIMD macs outside the loop and 4 for every iteration of the loop
************************************************************************/

#include "defts101.h"
#include "asm_sprt_TS.h"

.global _block_fir;

/* program memory code */
//.section/pm seg_pmco;
.section program;

_block_fir:
	leaf_entry;

/****************************************************
*	Save various registers to stack
*	This is simply added to show the use of the macros
*****************************************************/	
	/* save all register block to stack */
	save_reg;
	
	/* save some other registers to stack */
	puts = j0;;
	puts = j5;;
	puts = k0;;
	puts = k5;;

/****************************************************
* 	Retrieve passed arguments
*	The 4 pointers are passed through J7:4
*	The number of samples and number of taps
*	are integer values passed on the stack
*	j4 = input
*	j5 = dline
*	j6 = coeffs
*	j7 = output
*	reads(1) = samples
*	reads(2) = TAPS
*****************************************************/	
	
	xr17 = reads(1);	// N number of samples
	xr16 = reads(2);	// Number of TAPS 
	
	/* dline pointer CB setup */
	/* second parameter passed stired in r8 */
/*	j0 = j5;; jb0 = j0;; jl0 = xr1;;*/	j0 = j5;; jb0 = j0;; jl0 = xr16;;

/* Create new pointer in dline for storing new input */
	j1 = j0;; jb1 = j0;; jl1 = xr16;;
	
	/* input samples buffer setup */
	/* first parameter passed stored in r4 */
	j5 = j4;;
	
	/* coeffs pointer CB setup */
	/* third parameter passed stored in r12 */
	k0 = j6;; kb0 = k0;; kl0 = xr16;;
	
	/* output buffer setup */
	k5 = j7;;
	
	/* Setup modify registers for arrays */ 
	j12 = 0;;
	j13 = 1;;
	j14 = -1;;
	j15 = 2;;
	k13 = 2;;
	k14 = 1;;
	
	/* Zero the first output result */
	xr8 = 0;;
	
/* New instructions to prime the DAB */
	

/* The following two instructions can be replaced with the priming of the DAB instructions */
	/* move pointer to delay[1] */
/*	yr0 = CB [j0+=j13];;*/	r3:0 = DAB q[j0+=0];;
	
	/* load s0 with the value of delay[1] for SIMD store, move pointer to delay[0] */					
/*	yr0 = CB [j0+=j14];;*/	r3:0 = DAB q[j0+=4];;
							r3:0 = DAB q[j0+=4];;

	/* r1 = taps/2 due to SIMD mode */
/*	xr1 = lshift r1 by -1;;*/	xr16 = lshift r16 by -1;;
	
	/* 3 macs outside of fir mac loop */				
/*	xr3 = 3;;*/	xr18 = 4;;
	
	/* r3 = taps/2 - 3 for fir mac loop counter */ 							
/*	xr3 = r1 - r3;;*/	xr18 = r16 - r18;;
						xr18 = lshift r18 by -1;;	// two macs in inner loop now

    /* read one sample from INPUT[i] */
	xr0 = [j5+=j13];;
    
    /* read 2 coeffs */										
	yxr4 = CB l[k0+=k13];;
   	
   	/* outer loop - sample loop */					
/*	LC0 = xr2;; */	LC0 = xr17;;

/********************************************************************************************************
* In the following section of code the delay line is not always aligned on a dual address so we are 
* unable to perform SIMD accesses on the TigerSHARC. There are many ways to overcome this problem, 
* such as copying the delay line to another memory block then using a J and K pointer to load two 
* words in a single instruction. Another method would be to implement quad loads with the use of the
* data alignment buffer on TigerSHARC. In this original conversion we simply perform two loads instead
* of a SIMD load for the delay line with each having a modify of 1 instead of 2. A SIMD load however is
* performed for all coefficient accesses as the pointer in this buffer is always aligned.
********************************************************************************************************/		
		/* this can be modified as we only need to store the new sample on the TigerSHARC */
		/* This is due to the face that we do not have to perform a SIMD load or store as */
		/* there is no SIMD enable mode in TigerSHARC */
		/* transfer new sample and last new sample to delayline */
main_fir:	/*CB [j0+=j13] = xr0;;*/	CB [j1+=j14] = xr0;;
			/*CB [j0+=j13] = yr0;;*/
		
		/* samples * coeffs, read 2 samples, read 2 coeffs */			
/*		fr8 = r0*r4; xr0 = CB [j0+=j13]; yxr4 = CB l[k0+=k13];;*/ xfr8 = r0*r4; yfr8 = r1*r4; yxr4 = CB l[k0+=k13];;
/*		yr0 = CB [j0+=j13];;*/
		
		/* samples * coeffs, read 2 samples, read 2 coeffs */	
/*		fr12 = r0*r4; xr0 = CB [j0+=j13]; yxr4 = CB l[k0+=k13];;*/ xfr12 = r2*r4; yfr12 = r3*r4; yxr4 = CB l[k0+=k13];;
/*		yr0 = CB [j0+=j13];;*/ r3:0 = DAB q[j0+=4];;
  
		/* FIR loop */
/*		LC1 = xr3;;*/ LC1 = xr18;;

		/* samples * coeffs, accum, read 2 samples, read 2 coeffs */			
macs:   xfr12 = r0*r4; yfr12 = r1*r4; fr8 = r8+r12; yxr4 = CB l[k0+=k13];;
		xfr12 = r2*r4; yfr12 = r3*r4; fr8 = r8+r12; yxr4 = CB l[k0+=k13];;
		if NLC1E, jump macs;r3:0 = DAB q[j0+=4];; 
        
		/* samples * coeffs, accum, load s0 with the value of delay[1] for SIMD store, move pointer to delay[0] */
/*		fr12 = r0*r4; fr8 = r8+r12; yr0 = CB [j0+=j14];;*/	xfr12 = r0*r4; yfr12 = r1*r4; fr8 = r8+r12; yxr4 = CB l[k0+=k13];;
		
		/* new instruction as we need to perform 2 SIMD outside the loop due to quad operation */
		xfr12 = r2*r4; yfr12 = r3*r4; fr8 = r8+r12;;
        
		/* final SIMD accum, read new sample into s10 */
		fr8 = r8+r12; yr10 = [j5+=j13];;
        
		/* move PEy total into PEx register file */
		xr12 = yr8;;
		
		/* last accum, read sample into s0 for first MAC of next sample, read 2 coeffs */
		fr8 = r8+r12; yxr4 = CB l[k0+=k13];;

		/* move pointer for next round around the dline */
		j0 = j0 - 5 (cb);;
		
		/* prime the DAB again */
		r3:0 = DAB q[j0+=0];;
		r3:0 = DAB q[j0+=4];;
		r3:0 = DAB q[j0+=4];;
		
		/* place new sample from s10 into f0 without corrupting s0, swaps f0 and s10 only */
		xr0 = yr10;;
		
		
		/* write result to OUTPUT[i] */
		if NLC0E, jump main_fir; [k5+=k14] = xr8;;

	
	/* restore registers from stack */
	k5 = gets(1);
	k0 = gets(2);
	j5 = gets(3);
	j0 = gets(4);
	
	/* modify stack pointer */
	alter(16);
	
	/* restore register block from stack */
	restore_reg;
		
_block_fir.END:	leaf_exit;
